use crate::util::error::Error;
use crate::protocol::specs::RunMode;
use std::path;

#[derive(PartialEq, Eq, Debug)]
pub enum DataType {
    TEXT,
    FILE,
    DIR,
}

#[derive(Debug)]
pub struct Args {
    pub help: bool,
    pub version: bool,
    pub verbose_output: bool,
    pub mode: RunMode,
    pub send_type: DataType,
    pub address_str: String,
    pub send_path: path::PathBuf,
    pub save_directory: path::PathBuf,
    pub sync_directory: path::PathBuf,
    pub text_to_send: String,
}

impl Args {
    pub fn default() -> Args {
        return Args{
            help: false,
            version: false,
            verbose_output: false,
            mode: RunMode::RECEIVE,
            send_type: DataType::TEXT,
            address_str: "".to_string(),
            send_path: path::PathBuf::new(),
            save_directory: path::PathBuf::new(),
            sync_directory: path::PathBuf::new(),
            text_to_send: "".to_string(),
        };
    }
}

pub fn parse(raw_args: &Vec<String>) -> Result<Args, Error> {

    let mut parsed_args: Args = Args::default();

    // parse arguments, ignoring the first argument
    let arg_amount: usize = raw_args.len();
    if arg_amount < 2 {
        return Err(Error::new("no arguments"));
    }

    
    let mut arg_index: usize = 1;
    loop {
        match raw_args[arg_index].to_lowercase().as_str() {
            "help" => {
                parsed_args.help = true;
                break;
            }
    
            "version" => {
                parsed_args.version = true;
                break;
            }
    
            "verbose" => {
                parsed_args.verbose_output = true;
                arg_index += 1;
                continue;
            }
    
            "send" => {
                if arg_amount < 5 && !parsed_args.verbose_output {
                    // not enough arguments !
                    return Err(Error::new("not enough arguments for send mode"));
                } else if arg_amount < 6 && parsed_args.verbose_output {
                    return Err(Error::new("not enough arguments for send mode"));
                }
    
                parsed_args.mode = RunMode::SEND;
    
                match parsed_args.verbose_output {
                    true => {
                        arg_index = 3;
                    }
    
                    false => {
                        arg_index = 2;
                    }
                }

                // DataType
                match raw_args[arg_index].to_lowercase().as_str() {
                    "text" => {
                        parsed_args.send_type = DataType::TEXT;
                        parsed_args.text_to_send = String::from(&raw_args[arg_index + 1]);
                    }
    
                    "file" => {
                        parsed_args.send_type = DataType::FILE;
                        parsed_args.send_path = std::path::Path::new(&raw_args[arg_index + 1]).to_path_buf();
                    }
    
                    "dir" => {
                        parsed_args.send_type = DataType::DIR;
                        parsed_args.send_path = std::path::Path::new(&raw_args[arg_index + 1]).to_path_buf();
                    }
    
                    _ => {
                        return Err(Error::new(format!("invalid DataType {}", raw_args[arg_index]).as_str()));
                    }
                }
    
                // addr
                arg_index += 1;
                parsed_args.address_str = raw_args[arg_index + 1].to_string();

                break;
            }
    
            "recv" => {
                parsed_args.mode = RunMode::RECEIVE;            
                
                let arg_index: usize;
                match parsed_args.verbose_output {
                    true => {
                        arg_index = 3;
                    }
    
                    false => {
                        arg_index = 2;
                    }
                }
    
                if arg_amount > 2 && !parsed_args.verbose_output {
                    parsed_args.save_directory = path::Path::new(raw_args[arg_index].as_str()).to_path_buf();
                } else if arg_amount > 3 && parsed_args.verbose_output {
                    parsed_args.save_directory = path::Path::new(raw_args[arg_index].as_str()).to_path_buf();
                }

                break;
            }
    
            _ => {
                return Err(Error::new(format!("invalid argument {}", raw_args[arg_index]).as_str()));
            }
        }
    }

    return Ok(parsed_args);
}