use crate::util::error::Error;
use std::io::SeekFrom;
use std::io::Seek;
use std::io::Write;
use std::io::Read;

pub struct ChunkedFile {
    pub size: u128,
    pub chunk_size: u128,
    pub chunks_amount: u128,
    pub path: std::path::PathBuf,
}

impl ChunkedFile {
    fn empty() -> ChunkedFile {
        return ChunkedFile{
            size: 0,
            chunk_size: 0,
            chunks_amount: 0,
            path: std::path::PathBuf::new(),
        }
    }

    /// Create a new, ready-to-write chunked regular file
    pub fn new(path: &std::path::Path, size: u128, chunk_size: u128) -> Result<ChunkedFile, Error> {
        match std::fs::File::options().truncate(false).write(true).create(true).open(path) {
            Ok(_) => {}
        
            Err(error) => {
                return Err(
                    Error::new(format!("{}", error).as_str())
                );
            }
        }

        let mut new_chunked_file: ChunkedFile = ChunkedFile::empty();

        // path
        new_chunked_file.path = path.to_path_buf();

        // resize file 
        // match file_handle.set_len(size as u64) {
        //     Ok(()) => {}
        //     Err(error) => {
        //         return Err(
        //             Error::new(format!("could not resize \'{}\': {}", path.display(), error).as_str())
        //         );
        //     }
        // }
        new_chunked_file.size = size;

        // chunks
        new_chunked_file.chunk_size = chunk_size;
        if new_chunked_file.size % chunk_size != 0 {
            new_chunked_file.chunks_amount = (new_chunked_file.size / chunk_size) + 1;
        } else {
            new_chunked_file.chunks_amount = new_chunked_file.size / chunk_size;
        }

        return Ok(new_chunked_file);
    }

    /// Create a new, ready-to-read chunked file from already existing regular file
    pub fn from(path: &std::path::Path, chunk_size: u128) -> Result<ChunkedFile, Error> {
        if !path.is_file() {
            return Err(
                Error::new(format!("\'{}\' does not exist or not a regular file", path.display()).as_str())
            );
        }
        
        let mut new_chunked_file: ChunkedFile = ChunkedFile::empty();
        
        // path
        new_chunked_file.path = path.to_path_buf();

        // size
        match path.metadata() {
            Ok(metadata) => {
                new_chunked_file.size = metadata.len() as u128;
            }
            Err(error) => {
                return Err(
                    Error::new(format!("could not retrieve metadata for \'{}\': {}", path.display(), error).as_str())
                );
            }
        }

        // chunks
        new_chunked_file.chunk_size = chunk_size;
        if new_chunked_file.size % chunk_size != 0 {
            new_chunked_file.chunks_amount = (new_chunked_file.size / chunk_size) + 1;
        } else {
            new_chunked_file.chunks_amount = new_chunked_file.size / chunk_size;
        }

        return Ok(new_chunked_file);
    }

    pub fn read_chunk(&self, chunk_no: u128) -> Result<Vec<u8>, Error> {
        if chunk_no + 1 > self.chunks_amount {
            return Err(
                Error::new(format!("chunk index ({}) is bigger than there are chunks ({})", chunk_no, self.chunks_amount).as_str())
            );
        }
        
        // open underlying file
        let mut file_handle: std::fs::File;
        match std::fs::File::open(&self.path) {
            Ok(fh) => {
                file_handle = fh;
            }
            Err(error) => {
                return Err(Error::new(format!("{}", error).as_str()));
            }
        }
        
        // set position and read
        let chunk_start_pos: u128 = self.chunk_size * chunk_no;
        match file_handle.seek(SeekFrom::Start(chunk_start_pos as u64)) {
            Ok(_) => {}
            Err(error) => {
                return Err(Error::new(format!("seek error: {}", error).as_str()))
            }
        }

        let bytes_to_read: u128;
        if chunk_start_pos + self.chunk_size > self.size {
            bytes_to_read = chunk_start_pos + self.chunk_size - self.size;
        } else {
            bytes_to_read = self.chunk_size;
        }

        // read call leaves buf empty if it is not pre-initialized beforehand
        let mut buf: Vec<u8> = vec!(0; bytes_to_read as usize);
        match file_handle.read_exact(&mut buf) {
            Ok(_) => {}
            Err(error) => {
                return Err(Error::new(format!("read error: {}", error).as_str()))
            }
        }

        return Ok(Vec::from(buf));
    }

    pub fn write_chunk(&mut self, chunk: &[u8], chunk_no: u128) -> Result<(),Error> {
        if chunk_no + 1 > self.chunks_amount {
            // can't write outsize of pre-determined boundaries of a file !
            return Err(
                Error::new(format!("chunk index ({}) is bigger than there are chunks ({})", chunk_no, self.chunks_amount).as_str())
            );
        }

        if (chunk.len() as u128) < self.chunk_size && chunk_no + 1 != self.chunks_amount {
            // unfull chunk in the middle of a file
            return Err(Error::new("cannot write unfull chunk in a not last position"));
        }

        if (chunk.len() as u128) > self.chunk_size {
            // too big chunk
            return Err(Error::new("chunk is too big"));
        }

        // open file and write a new chunk
        let mut file_handle: std::fs::File;
        match std::fs::File::options().truncate(false).write(true).create(true).open(&self.path) {
            Ok(file) => {
                file_handle = file;
            }

            Err(error) => {
                return Err(Error::new(format!("{}", error).as_str()));
            }
        }

        let chunk_start_pos: u128 = self.chunk_size * chunk_no;
        match file_handle.seek(SeekFrom::Start(chunk_start_pos as u64)) {
            Ok(_) => {}
            Err(error) => {
                return Err(Error::new(format!("seek error: {}", error).as_str()));
            }
        }
        
        match file_handle.write_all(chunk) {
            Ok(()) => {}
            Err(error) => {
                return Err(Error::new(format!("write error: {}", error).as_str()))
            }
        }

        return Ok(());
    }
}

#[cfg(test)]
mod tests {
    use crate::fsys::file::ChunkedFile;

    #[test]
    fn test_file_operations() {
        let mut test_contents: Vec<u8> = vec!(0; 250);
        for i in 0..250 {
            test_contents[i] = i as u8;
        }

        {
            let mut new_file: ChunkedFile;
            match ChunkedFile::new(std::path::Path::new("testfile"), 255, 10) {
                Ok(f) => {
                    new_file = f;
                }
    
                Err(error) => {
                    panic!("{}", error.text);
                }
            }
            
            for i in 0..250/10 {
                if let Err(error) = new_file.write_chunk(&test_contents[i*10..i*10+10], i as u128) {
                    panic!("{}", error.text);
                }
            }
        }

        let chunked_file: ChunkedFile;
        match  ChunkedFile::from(std::path::Path::new("testfile"), 10) {
            Ok(f) => {
                chunked_file = f;
            }

            Err(error) => {
                panic!("{}", error.text);
            }
        }

        let mut read_contents: Vec<u8> = Vec::new();
        for i in 0..250/10 {
            match chunked_file.read_chunk(i) {
                Ok(chunk) => {
                    read_contents.extend_from_slice(&chunk);
                }

                Err(error) => {
                    panic!("{}", error.text);                    
                }
            }
        }

        // panic!("{:?} ---- {:?}", read_contents, test_contents);

        for i in 0..250 {
            if read_contents[i] != test_contents[i] {
                panic!("contents do not match");
            }
        }
    }
}