mod util;
mod protocol;
mod args;
mod crypt;
mod fsys;
mod nodes;

use protocol::specs::RunMode;
use args::parser::Args;
use args::parser::DataType;
use nodes::recv_node::RecvOptions;
use nodes::send_node::SendOptions;

const VERSION: &str = "v0.1.0";
const HELP_MESSAGE: &str = 
"ddtu - digital data transferring tool v0.1.0
Usage:
ddtu [FLAG] [MODE] [MODE-specific arguments]

[FLAG]
[help]: ddtu help -> print help message
[version]: ddtu version -> print version
[verbose]: ddtu verbose [MODE] [MODE-specific arguments] -> print almost everything the program does

[MODE]
[send]: ddtu send [TYPE] [SRC] [ADDR] -> send [SRC] of [TYPE] to [ADDR]
[recv]: ddtu recv [DIR](optional) -> wait for connection and receive anything from another ddtu instance, saving it to [DIR]

[TYPE]
[text]: ddtu send text \'text to send\' -> send text
[file]: ddtu send file \'/home/user/documents/doc.txt\' -> send file doc.txt
[dir]: ddtu send dir \'/home/user/homework\' -> send whole directory recursively

(c) Kasyanov Nikolay Alexeyevich (Unbewohnte)
";

const DEFAULT_PORT: u16 = 55761; 


fn main() {
    // get and parse arguments
    let raw_args: Vec<String> = std::env::args().collect();
    let args: Args;

    match args::parser::parse(&raw_args) {
        Ok(parsed_args) => {
            args = parsed_args;
        }

        Err(error) => {
            println!("error parsing arguments: {}", error.text);
            return;
        }
    }
    
    // handle them
    if args.help {
        println!("{}", HELP_MESSAGE);
        return;
    }

    if args.version {
        println!("ddtu {}", VERSION);
        return;
    }

    match args.mode {
        RunMode::RECEIVE => {
            match util::local_ip::get_addr() {
                Ok(local_ip_addr) => {
                    println!("Listening on {}:{}", local_ip_addr.to_string(), DEFAULT_PORT);
                }

                Err(_) => {
                    println!("Listening on port {}", DEFAULT_PORT);
                }
            }

            match nodes::recv_node::start(
                &mut RecvOptions::new(format!("0.0.0.0:{}", DEFAULT_PORT), args.save_directory), args.verbose_output) {
                Ok(_) => {}
                Err(error) => {
                    println!("Error receiving: {}", error.text);
                    return;
                }
            }
        }

        RunMode::SEND => {
            match args.send_type {
                DataType::DIR => {
                    println!("Sending directory is not implemented");
                    return;
                }

                DataType::FILE => {
                    println!("Sending files is not implemented");
                    return;
                }

                _ => {}
            }

            match nodes::send_node::start(
                SendOptions::new(
                    args.address_str,
                    args.send_type,
                    args.text_to_send,
                    args.send_path,
                ), args.verbose_output)
                
                {

                    Ok(_) => {}
                    Err(error) => {
                        println!("Error sending: {}", error.text);
                        return;    
                    }
                }
        }
    }
}
