use crate::util::error::Error;
use crate::protocol::specs::Packet;
use crate::protocol::packets::{read_next_packet, send_packet};
use crate::protocol::packets;
use crate::protocol::specs;
// use crate::fsys;
// use std::collections::HashMap;
use std::path;
use std::net;
use std::sync::mpsc::{Sender, Receiver};
use std::sync::mpsc;
use std::thread;

pub struct RecvOptions {
    pub listen_address: String,
    pub save_directory: path::PathBuf,
}

impl RecvOptions {
    pub fn new(addr: String, save_dir: path::PathBuf) -> RecvOptions {
        return RecvOptions{
            listen_address: addr,
            save_directory: save_dir,
        }
    }
}


pub fn start(options: &mut RecvOptions, verbose_output: bool) -> Result<(), Error> {
    let listener: net::TcpListener;
    match net::TcpListener::bind(&options.listen_address) {
        Ok(l) => {
            listener = l;
        }
        Err(error) => {
            return Err(Error::new(format!("could not create TCP listener: {}", error).as_str()));
        }
    }


    // wait for incoming connection from another ddtu instance,
    // reject any connection that doesn't provide handshake information
    let mut connection: net::TcpStream;
    let mut address: net::SocketAddr;
    let handshake: packets::Handshake;
    loop {
        match listener.accept() {
            Ok((c, a)) => {
                connection = c;
                address = a;
            }
            Err(error) => {
                return Err(Error::new(format!("error accepting a new incoming connection: {}", error).as_str()));
            }
        }
    
        if verbose_output {
            println!("Connection from {}", address);
        }

        // read the first packet
        match read_next_packet(&mut connection, specs::ENCRYPTION_NO_ENCRYPTION, &vec!(0)) {
            Ok(packet) => {
                match packets::Handshake::from_raw(&packet) {
                    Ok(h) => {
                        handshake = h;

                        if verbose_output {
                            println!("Handshake from {}", address);
                        }
        
                        // accept handshake
        
                        match send_packet(
                            &mut connection,
                            &packets::HandshakeAccept{}.as_raw(),
                            handshake.encryption_type,
                            &handshake.encryption_key
                        ) {
                            Ok(()) => {}
                            Err(error) => {
                                packets::close_connection(&mut connection, handshake.encryption_type, &handshake.encryption_key);
                                return Err(
                                    Error::new(format!("error accepting handshake from {}: {}", address, error.text).as_str())
                                );
                            }
                        }
        
                        // break out of loop
                        break;
                    }

                    Err(error) => {
                        packets::close_connection(&mut connection, specs::ENCRYPTION_NO_ENCRYPTION, &vec!(0));
                        return Err(Error::new(format!("error constructing handshake from {}: {}", address, error.text).as_str()));
                    }
                }
            }

            Err(error) => {
                // not a valid packet; it's most probably not another instance
                if verbose_output {
                    println!("{} did not provide valid handshake packet: {}. Dropping connection...", address, error.text);
                }
                packets::close_connection(&mut connection, specs::ENCRYPTION_NO_ENCRYPTION, &vec!(0));
                continue;
            }
        }
    }


    // remember file IDs and paths
    // let mut monitored_files: HashMap<u64, packets::FileInfo> = HashMap::new();
    
    // create save directory
    match std::fs::create_dir_all(&options.save_directory) {
        Ok(_) => {}
        Err(error) => {
            if verbose_output {
                println!("Could not create save directory: {}; saving to current one", error);
            }
            options.save_directory = path::Path::new(".").to_path_buf();
        }
    }

    // launch a new packet reading thread
    let (ch_send, ch_recv): (Sender<Packet>, Receiver<Packet>) = mpsc::channel();
   
    let mut cloned_connection: net::TcpStream;
    match connection.try_clone() {
        Ok(cc) => {
            cloned_connection = cc;
        }
        Err(_) => {
            packets::close_connection(&mut connection, handshake.encryption_type, &handshake.encryption_key);
            return Err(
                Error::new("could not clone connection")
            );
        }
    }
    cloned_connection.set_read_timeout(Some(std::time::Duration::from_secs(5)));

    let handshake_clone: packets::Handshake = handshake.clone();
    thread::spawn(move || {
        loop {
            match read_next_packet(&mut cloned_connection, handshake_clone.encryption_type, &handshake_clone.encryption_key) {
                Ok(packet) => {
                    if let Err(_) = ch_send.send(packet) {
                        // the channel has been hung up
                        // return from this thread
                        return;
                    }
                }
    
                Err(error) => {
                    if verbose_output {
                        println!("Could not read packet: {}", error.text);
                    }
                    return;
                }
            }
        }
    });


    // handle incoming packets
    loop {
        let incoming_packet: Packet;
        match ch_recv.recv() {
            Ok(p) => {
                incoming_packet = p;
            }
            Err(error) => {
                packets::close_connection(&mut connection, handshake.encryption_type, &handshake.encryption_key);
                return Err(Error::new(
                    format!("error receiving a new packet from listener thread: {}", error).as_str()
                ));
            }
        }

        match incoming_packet.header {
            specs::CONNECTION_SHUTDOWN_PACKET_ID => {
                if verbose_output {
                    println!("Connection has been shut down");
                }
                packets::close_connection(&mut connection, handshake.encryption_type, &handshake.encryption_key);
                
                return Ok(());
            }

            specs::TEXT_PACKET_ID => {
                match packets::Text::from_raw(&incoming_packet) {
                    Ok(text_packet) => {
                        // print received text !
                        println!("\n{}\n", text_packet.text);
                    }

                    Err(error) => {
                        // close connection and exit
                        packets::close_connection(&mut connection, handshake.encryption_type, &handshake.encryption_key);
                        return Err(
                            Error::new(format!("could not get text packet: {}", error.text).as_str())
                        );
                    }
                }
            }

            // specs::FILEINFO_PACKET_ID => {
            //     match packets::FileInfo::from_raw(&incoming_packet) {
            //         Ok(mut fileinfo_packet) => {                  
            //             // create it
            //             if let Err(error) = std::fs::File::create(&options.save_directory.join(&fileinfo_packet.relative_path)) {                
            //                 packets::close_connection(&mut connection, handshake.encryption_type, &handshake.encryption_key);
            //                 return Err(
            //                     Error::new(
            //                         format!("could not create a new file {}: {}", fileinfo_packet.filename, error).as_str()
            //                     ),
            //                 )
            //             }

            //             // accept it
            //             if let Err(error) = send_packet(
            //                 &mut connection,
            //                 &packets::FileAccept{}.as_raw(),
            //                 handshake.encryption_type,
            //                 &handshake.encryption_key,
            //             ) {
            //                 packets::close_connection(&mut connection, handshake.encryption_type, &handshake.encryption_key);
            //                 return Err(
            //                     Error::new(format!("could not accept file {}: {}", fileinfo_packet.filename, error.text).as_str()),
            //                 )
            //             }

            //             println!("\nReceiving \"{}\" ({} MB))", fileinfo_packet.filename, fileinfo_packet.filesize as f32/1024.0/1024.0);

            //             // monitor
            //             fileinfo_packet.relative_path = options.save_directory.join(fileinfo_packet.relative_path).to_string_lossy().into_owned();
            //             monitored_files.insert(
            //                 fileinfo_packet.file_id, fileinfo_packet,
            //             );
            //         }

            //         Err(error) => {
            //             packets::close_connection(&mut connection, handshake.encryption_type, &handshake.encryption_key);
            //             return Err(Error::new(format!("error getting a fileinfo packet: {}", error.text).as_str()));
            //         }
            //     }
            // }

            // specs::FILEDATA_PACKET_ID => {
            //     let filedata_packet: packets::FileData;
            //     match packets::FileData::from_raw(&incoming_packet) {
            //         Ok(fd_packet) => {
            //             filedata_packet = fd_packet;
            //         }

            //         Err(error) => {
            //             packets::close_connection(&mut connection, handshake.encryption_type, &handshake.encryption_key);
            //             return Err(Error::new(format!("error getting a filedata packet: {}", error.text).as_str()));
            //         }
            //     }

            //     // check if this file indeed was notified about
            //     match monitored_files.get(&filedata_packet.file_id) {
            //         None => {
            //             packets::close_connection(&mut connection, handshake.encryption_type, &handshake.encryption_key);
            //             return Err(Error::new(format!("file with ID {} was not notified about", filedata_packet.file_id).as_str()));
            //         }

            //         Some(file_info) => {
            //             // open file
            //             let mut chunked_file: fsys::file::ChunkedFile;
            //             match fsys::file::ChunkedFile::new(&path::Path::new(&file_info.relative_path), file_info.filesize, handshake.chunk_size as u128) {
            //                 Ok(f) => {
            //                     chunked_file = f;
            //                 }

            //                 Err(error) => {
            //                     packets::close_connection(&mut connection, handshake.encryption_type, &handshake.encryption_key);
            //                     return Err(
            //                         Error::new(format!("error opening chunked file {}: {}", file_info.filename, error.text).as_str())
            //                     );
            //                 }
            //             }
            //             // write required chunk
            //             if let Err(error) = chunked_file.write_chunk(&filedata_packet.chunk, filedata_packet.chunk_no as u128) {
            //                 packets::close_connection(&mut connection, handshake.encryption_type, &handshake.encryption_key);
            //                 return Err(
            //                     Error::new(format!("error writing a chunk to {}: {}", file_info.filename, error.text).as_str()),
            //                 );
            //             }
            //         }
            //     } 
            // }

            _ => {}
        }
    }
}
