use crate::util::error::Error;
use crate::args::parser::DataType;
use crate::crypt;
use crate::protocol::packets;
use crate::protocol::packets::{send_packet, read_next_packet};
use crate::protocol::specs;
use crate::protocol::specs::Packet;
use crate::fsys;
use std::thread;
use std::sync::mpsc::{Sender, Receiver};
use std::sync::mpsc;
use std::net;
use std::path;

pub struct SendOptions {
    pub address: String,
    pub send_type: DataType,
    pub text_to_send: String,
    pub source_path: path::PathBuf,
}

impl SendOptions {
    pub fn new(addr: String, send_t: DataType, txt_to_send: String, src_path: path::PathBuf) -> SendOptions {
        return SendOptions{
            address: addr,
            send_type: send_t,
            text_to_send: txt_to_send,
            source_path: src_path,
        };
    }
}

pub fn start(options: SendOptions, verbose_output: bool) -> Result<(), Error> {
    // connect
    let mut connection: net::TcpStream;
    match net::TcpStream::connect(&options.address) {
        Ok(conn) => {
            connection = conn;
        }

        Err(error) => {
            return Err(
                Error::new(format!("connection error: {}", error).as_str())
            );
        }
    }

    if verbose_output {
        println!("Connected to {}", &options.address);
    }

    // send handshake
    let handshake: packets::Handshake = packets::Handshake{
        protocol_version: specs::PROTOCOL_VERSION_1,
        chunk_size: specs::CHUNK_SIZE as u64,
        encryption_type: specs::ECRYPTION_XOR,
        encryption_key: crypt::keys::Key8bit::new_random().as_bytes(),
    };

    if verbose_output {
        println!("Using encryption ID: {} with key: {:?}", handshake.encryption_type, handshake.encryption_key);
    }
    
    match send_packet(&mut connection, &handshake.as_raw(), specs::ENCRYPTION_NO_ENCRYPTION, &vec!(0)) {
        Ok(_) => {
            if verbose_output {
                println!("Sent handshake");
            }
        }
        
        Err(error) => {
            packets::close_connection(&mut connection, specs::ENCRYPTION_NO_ENCRYPTION, &vec!(0));
            return Err(
                Error::new(format!("could not send handshake: {}", error.text).as_str())
            );
        }
    }

    // see if the handshake has been accepted
    match read_next_packet(&mut connection, handshake.encryption_type, &handshake.encryption_key) {
        Ok(packet) => {
            if packet.header != specs::HANDSHAKE_ACCEPT_PACKET_ID {
                // not approved or entirely not according to the protocol
                packets::close_connection(&mut connection, specs::ENCRYPTION_NO_ENCRYPTION, &vec!(0));
                return Err(
                    Error::new(format!("handshake hasn't been approved").as_str())
                );
            }

            if verbose_output {
                println!("Handshake has been accepted");
            }
        }

        Err(error) => {
            packets::close_connection(&mut connection, specs::ENCRYPTION_NO_ENCRYPTION, &vec!(0));
            return Err(
                Error::new(format!("could not receive potential handshake approval: {}", error.text).as_str())
            );
        }
    }

    // handshake has been accepted    
    // now send
    match options.send_type {
        DataType::TEXT => {
            if let Err(error) = send_packet(
                &mut connection,
                &packets::Text::new(&options.text_to_send).as_raw(),
                handshake.encryption_type,
                &handshake.encryption_key
            ) {
                packets::close_connection(&mut connection, handshake.encryption_type, &handshake.encryption_key);
                return Err(Error::new(format!("error sending text: {}", error.text).as_str()))
            }
            packets::close_connection(&mut connection, handshake.encryption_type, &handshake.encryption_key);
        }

        DataType::FILE => {
        //     // prepare file
        //     let chunked_file: fsys::file::ChunkedFile;
        //     match fsys::file::ChunkedFile::from(&options.source_path, handshake.chunk_size as u128) {
        //         Ok(f) => {
        //             chunked_file = f;
        //         }

        //         Err(error) => {
        //             packets::close_connection(&mut connection, handshake.encryption_type, &handshake.encryption_key);
        //             return Err(
        //                 Error::new(
        //                     format!("error constructing a chunked file {}: {}", options.source_path.display(), error.text).as_str())
        //                 );
        //         }
        //     }

        //     // send fileinfo packet
        //     let fileinfo_packet: packets::FileInfo;
        //     match packets::FileInfo::new(&chunked_file, 1, "".to_string()) {
        //         Ok(fp) => {
        //             fileinfo_packet = fp;
        //         }

        //         Err(error) => {
        //             packets::close_connection(&mut connection, handshake.encryption_type, &handshake.encryption_key);
        //             return Err(
        //                 Error::new(
        //                     format!("error constructing fileinfo packet for {}: {}", options.source_path.display(), error.text).as_str())
        //                 );
        //         }
        //     }

        //     match send_packet(&mut connection, &fileinfo_packet.as_raw(), handshake.encryption_type, &handshake.encryption_key) {
        //         Ok(()) => {}
                
        //         Err(error) => {
        //             packets::close_connection(&mut connection, handshake.encryption_type, &handshake.encryption_key);
        //             return Err(
        //                 Error::new(
        //                     format!("could not send fileinfo packet for {}: {}", options.source_path.display(), error.text).as_str())
        //                 );
        //         }
        //     }

        //     // wait for acceptance or rejection
        //     match read_next_packet(&mut connection, handshake.encryption_type, &handshake.encryption_key) {
        //         Ok(packet) => {
        //             match packet.header {
        //                 specs::FILE_ACCEPT_PACKET_ID => {
        //                     // proceed
        //                     if verbose_output {
        //                         println!("File has been accepted");
        //                     }
        //                 }

        //                 specs::FILE_REJECT_PACKET_ID | _ => {
        //                     // file not needed
        //                     packets::close_connection(&mut connection, handshake.encryption_type, &handshake.encryption_key);
        //                     return Ok(());
        //                 }
        //             }
        //         }

        //         Err(error) => {
        //             packets::close_connection(&mut connection, handshake.encryption_type, &handshake.encryption_key);
        //             return Err(
        //                 Error::new(
        //                     format!("could not receive answer for fileinfo packet: {}", error.text).as_str())
        //                 );
        //         }
        //     }

        //     println!("started to send chunks");
        //     // send the file in chunks !
        //     for chunk_no in 0..chunked_file.chunks_amount {
        //         // read next chunk
        //         let chunk_bytes: Vec<u8>;
        //         match chunked_file.read_chunk(chunk_no) {
        //             Ok(c) => {
        //                 chunk_bytes = c;
        //             }

        //             Err(error) => {
        //                 packets::close_connection(&mut connection, handshake.encryption_type, &handshake.encryption_key);
        //                 return Err(
        //                     Error::new(format!("error reading chunk no {} from {}: {}", chunk_no, chunked_file.path.display(), error.text).as_str()),
        //                 );
        //             }
        //         }

        //         // send it
        //         let filedata_packet: packets::FileData = packets::FileData::new(fileinfo_packet.file_id, chunk_no as u64, chunk_bytes);
        //         match send_packet(&mut connection, &filedata_packet.as_raw(), handshake.encryption_type, &handshake.encryption_key) {
        //             Ok(()) => {}
        //             Err(error) => {
        //                 packets::close_connection(&mut connection, handshake.encryption_type, &handshake.encryption_key);
        //                 return Err(
        //                     Error::new(
        //                         format!("could not send filedata packet for {}: {}", options.source_path.display(), error.text).as_str())
        //                     );
        //             }
        //         }
        //     }

        //     // once the file has been received - close the connection
        //     packets::close_connection(&mut connection, handshake.encryption_type, &handshake.encryption_key);
        //     return Ok(());
        }

        DataType::DIR => {

        }
    }

    return Ok(());
}
