use std::net;
use std::io::Write;
use std::io::Read;

use crate::util::error::Error;
use crate::util::buf_read::{read_u128_slice_be, read_utf8_string_slice, read_u64_slice_be};
use crate::protocol::specs::*;
use crate::fsys::file::ChunkedFile;
use crate::crypt;


pub fn send_packet<W>(conn: &mut W, packet: &Packet, encr_type: u8, key: &Vec<u8>) -> Result<(), Error>
 where W: Write {
    // ready-to-transport packet bytes
    let mut packet_bytes: Vec<u8> = Vec::with_capacity(packet.contents.len() + 1);
    packet_bytes.extend_from_slice(&packet.header.to_be_bytes());
    packet_bytes.extend_from_slice(&packet.contents);

    match encr_type {
        ECRYPTION_XOR => {
            crypt::xor::encrypt(&mut packet_bytes, &crypt::keys::Key8bit::new(key[0]));
        }

        ENCRYPTION_NO_ENCRYPTION | _ => {}
    }


    // get total length and add it to the beginning
    let total_packet_len: u64 = packet_bytes.len() as u64;

    let mut transport_bytes: Vec<u8> = Vec::with_capacity((total_packet_len + 8) as usize);
    transport_bytes.extend_from_slice(&total_packet_len.to_be_bytes());
    transport_bytes.append(&mut packet_bytes);

    // send packet
    match conn.write_all(&transport_bytes) {
        Ok(()) => {
            return Ok(());
        }

        Err(error) => {
            return Err(Error::new(format!("could not write packet to the connection: {}", error).as_str()))
        }
    }
}

pub fn read_next_packet<R>(conn: &mut R, encr_type: u8, key: &Vec<u8>) -> Result<Packet, Error>
 where R: Read {
    let mut u64_buf: [u8; 8] = [0; 8];
    loop {
        match conn.read_exact(&mut u64_buf) {
            Ok(()) => {
                break;
            }
    
            Err(error) => {
                // this is guaranteed to happen. It looks like networking in Rust is naive. Stream literally
                // is being read ONCE and that's it, even if it is empty, which results in an error.
                if error.kind() == std::io::ErrorKind::UnexpectedEof {
                    // try once again, maybe the packet has been delivered this time
                    continue;
                }

                return Err(
                    Error::new(format!("error reading initial length bytes: {}", error).as_str())
                );
            }
        }
    }

    let packet_len: u64 = u64::from_be_bytes(u64_buf);

    let mut packet_bytes: Vec<u8> = vec!(0; packet_len as usize);
    match conn.read_exact(&mut packet_bytes) {
        Ok(()) => {}
        Err(error) => {
            return Err(Error::new(format!("error while reading packet contents: {}", error).as_str()));
        }
    }

    if packet_bytes.len() < packet_len as usize {
        return Err(
            Error::new(
                format!("read {} packet bytes instead of specified {}", packet_bytes.len(), packet_len).as_str())
            );
    }

    // decrypt
    match encr_type {
        ECRYPTION_XOR => {
            crypt::xor::decrypt(&mut packet_bytes, &crypt::keys::Key8bit::new(key[0]));
        }

        ENCRYPTION_NO_ENCRYPTION | _ => {}
    }

    return Ok(
        Packet{
            header: packet_bytes[0].to_be(),
            contents: Vec::from(&packet_bytes[1..]),
        }
    );
}

pub fn close_connection(connection: &mut net::TcpStream, encr_type: u8, key: &Vec<u8>) {
    send_packet(connection, &ConnectionShutdown{}.as_raw(), encr_type, key);
    connection.shutdown(net::Shutdown::Both);
}

pub struct Text {
    pub text: String,
}

impl Text {
    pub fn new(txt: &str) -> Text {
        return Text {
            text: txt.to_string(),
        }
    }

    pub fn as_raw(&self) -> Packet {
        let mut raw_packet: Packet = Packet{
            header: TEXT_PACKET_ID,
            contents: Vec::with_capacity(self.text.len()),
        };

        let text_length: u64 = self.text.len() as u64;
        raw_packet.contents.extend_from_slice(&text_length.to_be_bytes());
        raw_packet.contents.extend_from_slice(&self.text.as_bytes());

        return raw_packet;
    }

    pub fn from_raw(raw_packet: &Packet) -> Result<Text, Error> {
        if raw_packet.header != TEXT_PACKET_ID {
            return Err(Error::new("it is not a text packet"));
        }

        if raw_packet.contents.len() < 9 {
            return Err(
                Error::new(format!("{} bytes is too small for a text packet to be valid", raw_packet.contents.len()).as_str()
            ));
        }

        // get text length
        let text_length: u64 = read_u64_slice_be(&raw_packet.contents[0..8]);
        
        if text_length as usize > raw_packet.contents[8..8+text_length as usize].len() {
            return Err(
                Error::new(
                    format!(
                        "text length ({}) is bigger than provided packet bytes length ({})",
                         text_length,
                         raw_packet.contents[8..8+text_length as usize].len()
                        ).as_str()
                    )
                );
        } 

        // extract text
        let packet_text = read_utf8_string_slice(&raw_packet.contents[8..8+text_length as usize]);

        return Ok(
            Text{
                text: packet_text,
            }
        );
    }
}


pub struct FileInfo {
    pub file_id: u64,
    pub filename: String,
    pub filesize: u128,
    pub relative_path: String,
}

impl FileInfo {
    fn empty() -> FileInfo {
        return FileInfo{
            file_id: 0,
            filename: String::new(),
            filesize: 0,
            relative_path: String::new(),
        };
    }

    pub fn new(file: &ChunkedFile, file_id: u64, rel_path: String) -> Result<FileInfo, Error> {
        let mut new_fileinfo_packet: FileInfo = FileInfo::empty();
        
        // id
        new_fileinfo_packet.file_id = file_id;
        
        //filename
        match file.path.file_name() {
            Some(filename) => {
                match filename.to_owned().into_string() {
                    Ok(filename_string) => {
                        new_fileinfo_packet.filename = filename_string;
                    }
                    Err(filename_os_string) => {
                        return Err(
                            Error::new(format!("could not convert \'{:?}\' (OsString) to String", filename_os_string).as_str())
                        );
                    }
                }
            }
            None => {
                return Err(
                    Error::new(format!("could not get filename from \'{}\'", file.path.display()).as_str())
                );
            }
        }

        // size
        new_fileinfo_packet.filesize =  file.size;

        // rel path
        if rel_path == "" {
            new_fileinfo_packet.relative_path = new_fileinfo_packet.filename.clone();
        } else {
            new_fileinfo_packet.relative_path = rel_path;
        }

        return Ok(new_fileinfo_packet);
    }

    pub fn as_raw(&self) -> Packet {
        let mut raw_packet: Packet = Packet{
            header: FILEINFO_PACKET_ID,
            contents: Vec::<u8>::new(),
        };
        
        // file id
        raw_packet.contents.extend_from_slice(&self.file_id.to_be_bytes());

        // filename
        let filename_bytes_length: u64 = self.filename.len() as u64;
        raw_packet.contents.extend_from_slice(&filename_bytes_length.to_be_bytes());
        raw_packet.contents.extend_from_slice(&self.filename.as_bytes());

        // filesize
        raw_packet.contents.extend_from_slice(&self.filesize.to_be_bytes());

        // relative path
        let relative_path_bytes_length: u64 = self.relative_path.len() as u64;
        raw_packet.contents.extend_from_slice(&relative_path_bytes_length.to_be_bytes());
        raw_packet.contents.extend_from_slice(&self.relative_path.as_bytes());

        return raw_packet;
    }

    pub fn from_raw(raw_packet: &Packet) -> Result<FileInfo, Error> {
        if raw_packet.header != FILEINFO_PACKET_ID {
            return Err(Error::new("it is not a fileinfo packet"));
        }

        if raw_packet.contents.len() < 8+8+16+8 {
            return Err(
                Error::new(format!("{} bytes is too small for a fileinfo packet to be valid", raw_packet.contents.len()).as_str()
            ));
        }

        // file id
        let file_id = read_u64_slice_be(&raw_packet.contents[0..8]);

        // get filename
        let filename_length: u64 = read_u64_slice_be(&raw_packet.contents[8..16]);
        let filename = read_utf8_string_slice(&raw_packet.contents[16..16+filename_length as usize]);

        // filesize
        let filesize = read_u128_slice_be(&raw_packet.contents[16+filename_length as usize..16+16+filename_length as usize]);

        // relative path
        let rel_path_length: u64 = read_u64_slice_be(
            &raw_packet.contents[32+filename_length as usize..32+8+filename_length as usize]
        );
        let relative_path = read_utf8_string_slice(
            &raw_packet.contents[32+8+filename_length as usize..(filename_length+32+8+rel_path_length) as usize]
        );

        return Ok(
            FileInfo{
                file_id: file_id,
                filename: filename,
                filesize: filesize,
                relative_path: relative_path,
            }
        );
    }
}


pub struct FileData {
    pub file_id: u64,
    pub chunk_no: u64,
    pub chunk: Vec<u8>,
}

impl FileData {
    pub fn new(file_id: u64, chunk_no: u64, chunk: Vec<u8>) -> FileData {
        return FileData{
            file_id: file_id,
            chunk_no: chunk_no,
            chunk: chunk,
        };
    }

    pub fn as_raw(&self) -> Packet {
        let mut raw_packet: Packet = Packet{
            header: FILEDATA_PACKET_ID,
            contents: Vec::<u8>::new(),
        };
        
        // file id
        raw_packet.contents.extend_from_slice(&self.file_id.to_be_bytes());

        // chunk number
        raw_packet.contents.extend_from_slice(&self.chunk_no.to_be_bytes());

        // chunk
        raw_packet.contents.extend_from_slice(&(self.chunk.len() as u64).to_be_bytes());
        raw_packet.contents.extend_from_slice(&self.chunk);

        return raw_packet;
    }

    pub fn from_raw(raw_packet: &Packet) -> Result<FileData, Error> {
        if raw_packet.header != FILEDATA_PACKET_ID {
            return Err(Error::new("it is not a filedata packet"));
        }

        if raw_packet.contents.len() < 8+8+8 {
            return Err(
                Error::new(format!("{} bytes is too small for a fileinfo packet to be valid", raw_packet.contents.len()).as_str()
            ));
        }

        // file id
        let file_id = read_u64_slice_be(&raw_packet.contents[0..8]);
        
        // chunk number
        let chunk_no = read_u64_slice_be(&raw_packet.contents[8..16]);

        // chunk bytes
        let chunk_length = read_u64_slice_be(&raw_packet.contents[16..24]);
        let mut chunk: Vec<u8> = vec!(0; chunk_length as usize);
        for i in 24..(24+chunk_length) as usize {
            chunk[i-24] = raw_packet.contents[i].to_be();
        }
        
        return Ok(
            FileData{
                file_id: file_id,
                chunk_no: chunk_no,
                chunk: chunk,
            }
        );
    }
}

pub struct ConnectionShutdown {}

impl ConnectionShutdown {
    pub fn as_raw(&self) -> Packet {
        return Packet{
            header: CONNECTION_SHUTDOWN_PACKET_ID,
            contents: Vec::new(),
        };
    }

    pub fn from_raw(raw_packet: &Packet) -> Result<ConnectionShutdown, Error> {
        match raw_packet.header {
            CONNECTION_SHUTDOWN_PACKET_ID => {
                return Ok(
                    ConnectionShutdown{},
                );
            }

            _ => {
                return Err(
                    Error::new("this is not a connection shutdown packet"),
                )
            }
        }
    }
}

#[derive(Clone, Debug)]
pub struct Handshake {
    pub protocol_version: u8,
    pub chunk_size: u64,
    pub encryption_type: u8,
    pub encryption_key: Vec<u8>,
}

impl Handshake {
    pub fn as_raw(&self) -> Packet {
        let mut raw_packet: Packet = Packet{
            header: HANDSHAKE_PACKET_ID,
            contents: Vec::<u8>::new(),
        };
        
        // protocol version
        raw_packet.contents.extend_from_slice(&PROTOCOL_LATEST_VERSION.to_be_bytes());

        // chunk_size
        raw_packet.contents.extend_from_slice(&u64::from_be_bytes(CHUNK_SIZE.to_be_bytes()).to_be_bytes());

        // encryption type
        raw_packet.contents.extend_from_slice(&self.encryption_type.to_be_bytes());

        // encryption key
        raw_packet.contents.extend_from_slice(&(self.encryption_key.len() as u64).to_be_bytes());
        raw_packet.contents.extend_from_slice(&self.encryption_key);

        return raw_packet;
    }

    pub fn from_raw(raw_packet: &Packet) -> Result<Handshake, Error> {
        if raw_packet.header != HANDSHAKE_PACKET_ID {
            return Err(Error::new("it is not a handshake packet"));
        }

        if raw_packet.contents.len() < 1+1+8+8 {
            return Err(
                Error::new(format!("{} bytes is too small for a handshake packet to be valid", raw_packet.contents.len()).as_str())
            );
        }

        // protocol version
        let protocol_version = raw_packet.contents[0].to_be();

        // chunk size
        let chunk_size = read_u64_slice_be(&raw_packet.contents[1..9]);

        // encryption type
        let encryption_type = raw_packet.contents[9].to_be();

        // key
        let key_len: u64 = read_u64_slice_be(&raw_packet.contents[10..18]);
        let mut encryption_key: Vec<u8> = vec!(0; key_len as usize);
        if key_len == 0 {
            encryption_key = Vec::new();
        } else {
            for i in 18..(18+key_len) as usize {
                encryption_key[i-18] = raw_packet.contents[i].to_be();
            }
        }

        return Ok(
            Handshake{
                protocol_version: protocol_version,
                chunk_size: chunk_size,
                encryption_type: encryption_type,
                encryption_key: encryption_key,
            }
        );
    }
}

pub struct HandshakeAccept {}

impl HandshakeAccept {
    pub fn as_raw(&self) -> Packet {
        return Packet{
            header: HANDSHAKE_ACCEPT_PACKET_ID,
            contents: Vec::new(),
        };
    }

    pub fn from_raw(raw_packet: &Packet) -> Result<HandshakeAccept, Error> {
        match raw_packet.header {
            HANDSHAKE_ACCEPT_PACKET_ID => {
                return Ok(HandshakeAccept{})
            }

            _ => {
                return Err(
                    Error::new("this is not a handshake accept packet"),
                )
            }
        }
    }
}

pub struct FileAccept {}

impl FileAccept {
    pub fn as_raw(&self) -> Packet {
        return Packet{
            header: FILE_ACCEPT_PACKET_ID,
            contents: Vec::new(),
        }
    }

    pub fn from_raw(raw_packet: &Packet) -> Result<FileAccept, Error> {
        match raw_packet.header {
            FILE_ACCEPT_PACKET_ID => {
                return Ok(FileAccept{})
            }

            _ => {
                return Err(
                    Error::new("this is not a file accept packet"),
                );
            }
        }
    }
}

pub struct FileReject {}

impl FileReject {
    pub fn as_raw(&self) -> Packet {
        return Packet{
            header: FILE_REJECT_PACKET_ID,
            contents: Vec::new(),
        }
    }

    pub fn from_raw(raw_packet: &Packet) -> Result<FileReject, Error> {
        match raw_packet.header {
            FILE_ACCEPT_PACKET_ID => {
                return Ok(FileReject{})
            }

            _ => {
                return Err(
                    Error::new("this is not a file reject packet"),
                );
            }
        }
    }
}